
#include <string.h>
#include <stdio.h>
#include <glib.h>
#include <gnet.h>
#include "http.h"

enum {
	HTTP_METHOD_LEN		= 8,
};

typedef void (*cli_evt_func)(GSHTTP_client *cli, GConnEvent *evt);

static void cli_st_get_req(GSHTTP_client *cli)
{
	cli->state = evt_get_req;
	gnet_conn_readline(cli->conn);
	gnet_conn_timeout(cli->conn, cli->gsh->idle_timeout);
}

static void cli_http_request(GSHTTP_client *cli)
{
	GSHTTP_event evt = { http_evt_req };
	evt.u.cli = cli;
	if (cli->gsh->func(cli->gsh, &evt, cli->gsh->user_data))
		cli_st_get_req(cli);
	else
		cli->state = evt_dispose;
}

static void req_free(struct http_req *req)
{
	if (!req)
		return;
	g_free(req->method);
	if (req->uri)
		gnet_uri_delete(req->uri);
	if (req->hdr)
		g_hash_table_unref(req->hdr);
	g_slice_free(struct http_req, req);
}

static struct http_req *req_new(void)
{
	struct http_req *req;

	req = g_slice_new0(struct http_req);
	if (!req)
		return NULL;

	req->hdr = g_hash_table_new_full(g_str_hash, g_str_equal,
					 g_free, g_free);
	return req;
}

static void cli_free(GSHTTP_client *cli)
{
	if (!cli)
		return;
	if (cli->conn)
		gnet_conn_unref(cli->conn);
	req_free(cli->req);
	if (cli->tmp_hdr)
		g_string_free(cli->tmp_hdr, TRUE);
	g_slice_free(GSHTTP_client, cli);
}

static void cli_err(GSHTTP_client *cli, const gchar *err_resp,
		    gboolean close_conn)
{
	gchar *resp;

	resp = g_strdup_printf("HTTP/%d.%d %s\r\n"
		"Content-type: text/html\r\n"
		"Connection: close\r\n"
		"\r\n"
		"<html><head><title>HTTP request error</title></head>"
		"<body><h1>HTTP request error</h1>"
		"<h2>Error: %s</h2></body></html>",
			       cli->req->major,
			       cli->req->minor,
			       err_resp, err_resp);

	gnet_conn_write(cli->conn, resp, strlen(resp));

	if (close_conn)
		cli->state = evt_dispose;
	else
		cli_st_get_req(cli);
}

static void cli_hdr_append(GHashTable *hdr, gchar *new_key, gchar *new_val,
			   gchar *old_val)
{
	gchar *tmp = g_strdup_printf("%s, %s", old_val, new_val);

	g_free(new_val);
	g_hash_table_remove(hdr, new_key);	/* frees old_key, old_val */
	g_hash_table_insert(hdr, new_key, tmp);
}

static gboolean cli_hdr_flush(GSHTTP_client *cli)
{
	struct http_req *req = cli->req;
	gchar *hdr, *tmp, *key, *val, *old_val;
	const gchar *err_resp;

	if (!cli->tmp_hdr)
		return TRUE;

	/* get accumulated header text */
	hdr = g_string_free(cli->tmp_hdr, FALSE);
	cli->tmp_hdr = NULL;

	/* find colon, make sure no whitespace comes before it */
	tmp = hdr;
	while (*tmp) {
		if (g_ascii_isspace(*tmp)) {
			err_resp = "400 whitespace in header key";
			goto err_out;
		}
		if (*tmp == ':')
			break;
		tmp++;
	}
	if (*tmp != ':') {
		err_resp = "400 header colon not found";
		goto err_out;
	}
	*tmp = 0;

	key = g_ascii_strdown(hdr, -1);

	tmp++;
	while ((*tmp) && (g_ascii_isspace(*tmp)))
		tmp++;

	val = g_strdup(tmp);

	/* if header already exists, append comma and value */
	old_val = g_hash_table_lookup(req->hdr, key);
	if (!old_val)
		g_hash_table_insert(req->hdr, key, val);
	else
		cli_hdr_append(req->hdr, key, val, old_val);

	g_free(hdr);
	return TRUE;

err_out:
	cli_err(cli, err_resp, TRUE);
	return FALSE;
}

static void cli_evt_dispose(GSHTTP_client *cli, GConnEvent *evt)
{
	cli_free(cli);
}

static void cli_evt_get_hdr(GSHTTP_client *cli, GConnEvent *evt)
{
	gchar *buf = evt->buffer;
	gint len = evt->length;
	gboolean eoh = FALSE;

	len--;			/* chop off \r\n */
	if ((len > 0) && (buf[len - 1] == '\r'))
		len--;
	buf[len] = 0;
	if (!len)		/* blank line signals end of headers */
		eoh = TRUE;

	/* check need to flush accumulated header data */
	if (eoh || (!g_ascii_isspace(buf[0]))) {
		gboolean sent_resp = !cli_hdr_flush(cli);
		if (sent_resp)
			return;
	}

	/* we have a fully formed HTTP request, deliver it */
	if (eoh)
		cli_http_request(cli);

	/* continue accumulating header data */
	else {
		if (!cli->tmp_hdr)
			cli->tmp_hdr = g_string_new_len(buf, len);
		else
			g_string_append_len(cli->tmp_hdr, buf, len);

		/* remain in this cli->state */
		gnet_conn_readline(cli->conn);
		gnet_conn_timeout(cli->conn, cli->gsh->idle_timeout);
	}
}

static void cli_evt_get_req(GSHTTP_client *cli, GConnEvent *evt)
{
	struct http_req *req = cli->req;
	gchar *sp1, *sp2, *buf = evt->buffer;
	gint len = evt->length;
	const gchar *err_resp;

	/* since we used gnet's readline, we know there is trailing
	 * whitespace, either \r\n or \n.  thus we can safely
	 * nul-terminate the string without losing anything useful
	 */

	buf[len] = 0;

#if 0
	/* hueristic: make us ASCII-sane, C-string-friendly,
	 * and friendly to all nations (we like utf8)
	 */
	if (!g_utf8_validate(buf, len, NULL)) {
		err_resp = "400 request string not UTF8";
		goto err_out;
	}
#endif

	/* locate the first and second spaces, additionally ensuring
	 * that the first and second tokens are non-empty
	 */
	if (*buf == ' ') {
		err_resp = "400 empty method";
		goto err_out;
	}
	sp1 = strchr(buf, ' ');
	if ((!sp1) || (*(sp1 + 1) == ' ')) {
		err_resp = "400 missing URI";
		goto err_out;
	}
	sp2 = strchr(sp1 + 1, ' ');
	if (!sp2) {
		err_resp = "505 missing HTTP version / version 0.9 unsupported";
		goto err_out;
	}

	/* convert the two spaces to nuls, thereby creating three
	 * nul-terminated strings for the three pieces we desire
	 */
	*sp1 = 0;
	*sp2 = 0;

	/* method is the first token, at the beginning of the buffer */
	req->method = g_ascii_strup(buf, -1);

	/* URI is the second token, immediately following the first space */
	req->uri = gnet_uri_new(sp1 + 1);
	if (!req->uri) {
		err_resp = "400 URI parse failed";
		goto err_out;
	}
	gnet_uri_unescape(req->uri);

	/* HTTP version is the final token, following second space */
	if ((sscanf(sp2 + 1, "HTTP/%d.%d", &req->major, &req->minor) != 2) ||
	    (req->major != 1) || (req->minor < 0) || (req->minor > 1)) {
		err_resp = "400 HTTP version parse failed";
		goto err_out;
	}

	cli->state = evt_get_hdr;
	gnet_conn_readline(cli->conn);
	gnet_conn_timeout(cli->conn, cli->gsh->idle_timeout);
	return;

err_out:
	cli_err(cli, err_resp, TRUE);
}

static cli_evt_func state_funcs[] = {
	[evt_get_req]	= cli_evt_get_req,
	[evt_get_hdr]	= cli_evt_get_hdr,
	[evt_dispose]	= cli_evt_dispose,
};

static void cli_event(GConn *conn, GConnEvent *evt, gpointer user_data)
{
	GSHTTP_client *cli = user_data;

	if ((evt->type == GNET_CONN_ERROR) || (evt->type == GNET_CONN_CLOSE) ||
	    (evt->type == GNET_CONN_TIMEOUT))
		cli->state = evt_dispose;

	state_funcs[cli->state](cli, evt);
}

static void gshttp_gsrv_event(GServer *gsrv, GConn *conn, gpointer user_data)
{
	GSHTTP *gsh = user_data;
	GSHTTP_client *cli;

	if (!conn) {
		GSHTTP_event evt = { http_evt_exit };
		gsh->func(gsh, &evt, gsh->user_data);
		return;
	}

	cli = g_slice_new0(GSHTTP_client);
	cli->gsh = gsh;
	cli->conn = conn;
	cli->req = req_new();
	cli->req->major = 1;
	cli->req->minor = 1;

	gnet_conn_set_callback(conn, cli_event, cli);

	cli_st_get_req(cli);
}

GSHTTP *gshttp_new(const GInetAddr *iface, gint port, GSHTTPFunc func,
		   gpointer user_data)
{
	GSHTTP *gsh;

	gsh = g_slice_new0(GSHTTP);
	if (!gsh)
		return NULL;

	gsh->ref = 1;
	gsh->idle_timeout = 15 * 60 * 1000;
	gsh->func = func;
	gsh->user_data = user_data;

	gsh->gsrv = gnet_server_new(iface, port, gshttp_gsrv_event, gsh);
	if (!gsh->gsrv)
		goto err_out;
	
	return gsh;

err_out:
	g_slice_free(GSHTTP, gsh);
	return NULL;
}

void gshttp_delete(GSHTTP *gsh)
{
	gnet_server_unref(gsh->gsrv);
	g_slice_free(GSHTTP, gsh);
}

void gshttp_unref(GSHTTP *gsh)
{
	gsh->ref--;
	if (!gsh->ref)
		gshttp_delete(gsh);
}

void gshttp_ref(GSHTTP *gsh)
{
	gsh->ref++;
}

